package net.sf.sido.factory.support;

import com.google.common.base.Predicate;
import net.sf.sido.*;
import net.sf.sido.array.Array;
import net.sf.sido.array.IndexedArray;
import net.sf.sido.constraint.ConstraintDefinitions;
import net.sf.sido.factory.DataFactoryReader;
import org.apache.commons.lang.StringUtils;

import java.io.IOException;
import java.net.URL;
import java.util.*;

public abstract class AbstractDataFactoryReader implements DataFactoryReader, PropertyAssemblerCallback {

    @Override
    public final DataSchema read(URL url, DataFactory factory) throws IOException {
        Collection<DataSchema> schemas = read(Collections.singleton(url), factory);
        // Assumes only one schema is returned
        return schemas.iterator().next();
    }

    @Override
    public final Collection<DataSchema> read(Collection<URL> urls, DataFactory factory) throws IOException {

        // Reads the SiDO schema
        SiDOSchema sidoSchema = new SiDOSchema();

        // Reads all meta-schemas
        Collection<DataObject> sSchemas = readSSchemas(urls, sidoSchema);

        // Parses the meta-schemas
        return readSchemas(factory, sSchemas);
    }

    protected Collection<DataSchema> readSchemas(DataFactory factory, Collection<DataObject> sSchemas) throws IOException {
        // Schemas to create and indexation of types
        Map<String,TypeDefinition> typeDefinitions = new LinkedHashMap<String,TypeDefinition>();
        Map<String, DataSchema> schemas = new LinkedHashMap<String, DataSchema>();
        for (DataObject sSchema : sSchemas) {
            String schemaName = sSchema.getString(SiDOSchema.NAME);
            // TODO Checks if the schema is already defined
            DataSchema schema = factory.createSchema(schemaName);
            schemas.put(schemaName, schema);
            for (DataObject sType : sSchema.getDOArray(SiDOSchema.TYPES)) {
                DataTypeBuilder builder = schema.createDataTypeBuilder(sType.getString(SiDOSchema.NAME));
                TypeDefinition typeDefinition = new TypeDefinition(schema, sType, builder);
                typeDefinitions.put(typeDefinition.getQualifiedName(), typeDefinition);
            }
        }

        // Reads all types
        readTypes(factory, schemas, typeDefinitions);

        // Extensions
        readExtensions(factory, sSchemas);

        // Closes all schemas
        for (DataSchema schema : schemas.values()) {
            schema.close();
        }

        // Returns the list of actual schemas
        return schemas.values();
    }

    protected Collection<DataObject> readSSchemas(Collection<URL> urls, SiDOSchema sidoSchema) throws IOException {
        // Reads all schemas
        Collection<DataObject> sSchemas = new ArrayList<DataObject>();
        for (URL url : urls) {
            DataObject sSchema = readSchema(sidoSchema, url);
            sSchemas.add(sSchema);
        }
        return sSchemas;
    }

    /**
     * All types are now ready, we just have to read and register all extensions.
     */
    protected void readExtensions(DataFactory factory, Collection<DataObject> sSchemas) {
        for (DataObject sSchema : sSchemas) {
            for (DataObject sExtension : sSchema.getDOArray(SiDOSchema.EXTENSIONS)) {
                readExtension(factory, sSchema, sExtension);
            }
        }
    }

    protected void readExtension(DataFactory factory, DataObject sSchema, DataObject sExtension) {
        // Gets the type reference
        String typeName = sExtension.getString(SiDOSchema.TYPEREF);
        // Gets the type (MUST be defined)
        DataType extendedType = factory.getMandatoryType(typeName);
        // For all extension properties
        for (DataObject property : sExtension.getDOArray(SiDOSchema.PROPERTIES)) {
            readExtensionProperty(factory, sSchema, extendedType, property);
        }
    }

    protected void readExtensionProperty(DataFactory factory, DataObject sSchema, DataType extendedType, DataObject property) {
		// Properties
		String name = property.getString(SiDOSchema.NAME);
		String type = property.getString(SiDOSchema.TYPEREF);
		// Additional properties
		boolean nullable = property.getBoolean(SiDOSchema.NULLABLE);
		boolean reverse = property.getBoolean(SiDOSchema.REVERSE);
        boolean array = property.getBoolean(SiDOSchema.ARRAY);
		String arrayIndexProperty = property.getString(SiDOSchema.ARRAY_INDEX);
		boolean arrayIndex = array && StringUtils.isNotBlank(arrayIndexProperty);
        boolean anonymous = property.getBoolean(SiDOSchema.ANONYMOUS);
		// Constraint
        Predicate<?> constraint = null;
        final Array<DataObject> constraints = property.getDOArray(SiDOSchema.CONSTRAINTS);
        if (constraints != null && !constraints.isEmpty()) {
            ConstraintDefinitions constraintDefinitions = new DefaultConstraintDefinitions(constraints);
            constraint = factory.getConstraintFactory().createConstraint(constraintDefinitions);
        }
        // Type information
        boolean simple = false;
        if (anonymous) {
            // Anonymous is not supported for extensions
            throw new IllegalStateException(String.format("Anonymous %s extension of type %s is not supported", name, extendedType.getQualifiedName()));
        } else {
            if (StringUtils.isBlank(type)) {
                type = "string";
            }
            simple = isSimpleType(factory, type);
            if (!simple) {
                if (!type.contains(DataSchema.QUALIFIER_SEPARATOR)) {
                    type = String.format("%s%s%s",
                            sSchema.getString(SiDOSchema.NAME),
                            DataSchema.QUALIFIER_SEPARATOR,
                            type
                            );
                }
            }
        }
		// Extends the type
		if (array && !simple) {
			// Gets the type in the current schema
			DataType propertyType = factory.getMandatoryType(type);
            if (arrayIndex) {
                extendedType.extendIndexedArray(name, propertyType, arrayIndexProperty, nullable, (Predicate<IndexedArray<DataObject,Object>>) constraint);
            } else {
                @SuppressWarnings("unchecked")
                Predicate<Array<DataObject>> doConstraint = (Predicate<Array<DataObject>>) constraint;
                extendedType.extendArray(name, propertyType, nullable, doConstraint);
            }
		} else if (array && simple) {
			Class<Object> simpleType = factory.getSimpleType(type);
			@SuppressWarnings("unchecked")
			Predicate<Array<Object>> simpleConstraint = (Predicate<Array<Object>>) constraint;
			extendedType.extendArray(name, simpleType, nullable, simpleConstraint);
		} else if (!simple) {
			// Gets the type in the current schema
			DataType propertyType = factory.getMandatoryType(type);
			@SuppressWarnings("unchecked")
			Predicate<DataObject> doConstraint = (Predicate<DataObject>) constraint;
			extendedType.extend(name, propertyType, nullable, reverse, doConstraint);
		} else {
			Class<Object> simpleType = factory.getSimpleType(type);
			@SuppressWarnings("unchecked")
			Predicate<Object> simpleConstraint = (Predicate<Object>) constraint;
			extendedType.extend(name, simpleType, nullable, simpleConstraint);
		}
    }

    protected void readTypes(DataFactory factory, Map<String, DataSchema> schemas, Map<String,TypeDefinition> typeDefinitions) throws IOException {
        for (TypeDefinition typeDefinition : typeDefinitions.values()) {
            TypeResolutionStatus status = typeDefinition.getStatus();
			if (status == TypeResolutionStatus.PENDING) {
				readType(factory, schemas, typeDefinitions, typeDefinition);
			} else if (status != TypeResolutionStatus.COMPLETE) {
				throw new IllegalStateException(String.format("Status %s for definition %s cannot be resolved.",
						status, typeDefinition.getQualifiedName()));
			}
        }
    }

    protected void readType(DataFactory factory, Map<String, DataSchema> schemas, Map<String,TypeDefinition> typeDefinitions, TypeDefinition definition) throws IOException {
		definition.start();
		// Super-type
		String parentTypeName = definition.getParentTypeName();
		if (StringUtils.isNotBlank(parentTypeName)) {
			// Gets the parent type reference
			DataType parentType = getDOType(factory, schemas, typeDefinitions, parentTypeName);
			// Sets as parent
			definition.setParentType(parentType);
		}
		// Gets all properties
        Array<DataObject> properties = definition.getProperties();
        for (DataObject property : properties) {
            readProperty(factory, schemas, typeDefinitions, definition, property);
        }
		// Completion
		definition.complete();
    }

    protected boolean isSimpleType (DataFactory factory, String typeName) {
        return StringUtils.isBlank(typeName) || (factory.forAlias(typeName) != null);
    }

    protected void readProperty(DataFactory factory, Map<String, DataSchema> schemas, Map<String,TypeDefinition> typeDefinitions, TypeDefinition definition, DataObject property) throws IOException {
		// Property name
		String name = property.getString(SiDOSchema.NAME);
        // Type attributes
        boolean anonymous = property.getBoolean(SiDOSchema.ANONYMOUS);
        boolean array = property.getBoolean(SiDOSchema.ARRAY);
		String arrayIndexProperty = property.getString(SiDOSchema.ARRAY_INDEX);
		boolean arrayIndex = array && StringUtils.isNotBlank(arrayIndexProperty);
		// Type name
        String typeName = null;
        boolean simple = false;
        if (!anonymous) {
            typeName = property.getString(SiDOSchema.TYPEREF);
            if (StringUtils.isBlank(typeName)) {
                typeName = "string";
            }
            simple = isSimpleType(factory, typeName);
        } else {
            simple = false;
        }
		// Creates the property assembler
		PropertyAssembler assembler;
		if (anonymous && !array) {
			// Anonymous type
			assembler = new AnonymousPropertyAssembler(name);
		} else if (anonymous && array) {
			// Array of generics
		    assembler = new AnonymousArrayPropertyAssembler(name);
		} else if (array && !simple) {
			if (arrayIndex) {
				assembler = new IndexedArrayDOPropertyAssembler(name, typeName, arrayIndexProperty);
			} else {
				assembler = new ArrayDOPropertyAssembler(name, typeName);
			}
		} else if (array && simple) {
			// Indexed array of simple properties is not allowed
			if (arrayIndex) {
				throw new IllegalStateException(
					String.format(
						"Indexed array of simple property %s for definition %s is not allowed.",
						name, definition.getQualifiedName()));
			}
			assembler = new ArraySimplePropertyAssembler(name, typeName);
		} else if (!simple) {
			assembler = new DOPropertyAssembler(name, typeName);
		} else {
			assembler = new SimplePropertyAssembler(name, typeName);
		}
		// Creates the property
		assembler.addProperty(factory, this, schemas, typeDefinitions, definition, property);
    }

    @Override
    public DataType getDOType(DataFactory factory, Map<String, DataSchema> schemas, Map<String,TypeDefinition> typeDefinitions, String qualifiedTypeName) throws IOException {
        // This type can be found from the current factory, not only in current schemas
        DataType type = factory.getType(qualifiedTypeName);
        if (type != null) {
            return type;
        }
		// Gets the definition for this type
		TypeDefinition tDefinition = typeDefinitions.get(qualifiedTypeName);
		if (tDefinition == null) {
			throw new DOTypeDefinitionNotFoundException(qualifiedTypeName);
		}
		// If not resolved, resolve it
		TypeResolutionStatus status = tDefinition.getStatus();
		if (status == TypeResolutionStatus.COMPLETE) {
			return tDefinition.getType();
		} else if (status == TypeResolutionStatus.PENDING) {
            readType(factory, schemas, typeDefinitions, tDefinition);
			return tDefinition.getType();
		} else if (status == TypeResolutionStatus.ONGOING) {
			return tDefinition.getBuilder().getType();
		} else {
			throw new IllegalStateException(String.format("Status %s is unknown.", status));
		}
    }
}
