package net.sf.sido.factory.support;

import net.sf.sido.*;
import net.sf.sido.support.DataFactories;

/**
 * Access to the schema of schemas.
 *
 * <pre>
schema Schema

// This schema defines the SiDO schema itself

a SSchema type
has
    a name,
    an array of SType as types,
    an array of SExtension as extensions
    .

a SType type
has
    a name,
    a boolean as xextensible,
    a boolean as xabstract,
    a nullable string as parent,
    an array of SProperty as properties
    .

a SProperty type
has
    a boolean as nullable,
    a boolean as reverse,
    a boolean as array,
    a nullable string as arrayIndex,
    a nullable boolean as anomnymous,
    a nullable string as typeref,
    a name,
    an array of SConstraint as constraints
    .

a SConstraint type
has
    an id,
    a nullable string as value,
    an array of SContraintParam as parameters
    .

a SContraintParam type
has
    a name,
    a value
    .

a SExtension type
has
    a typeref,
    an array of SProperty as properties
    .
</pre>
 */
public class SiDOSchema {

    public static final String EXTENSIONS = "extensions";
    public static final String TYPES = "types";
    public static final String NAME = "name";
    public static final String S_SCHEMA = "SSchema";
    public static final String PROPERTIES = "properties";
    public static final String TYPEREF = "typeref";
    public static final String S_EXTENSION = "SExtension";
    public static final String PARENT = "parent";
    public static final String ABSTRACT = "abstract";
    public static final String EXTENSIBLE = "extensible";
    public static final String S_TYPE = "SType";
    public static final String CONSTRAINTS = "constraints";
    public static final String ANONYMOUS = "anonymous";
    public static final String ARRAY = "array";
    public static final String ARRAY_INDEX = "arrayIndex";
    public static final String REVERSE = "reverse";
    public static final String NULLABLE = "nullable";
    public static final String S_PROPERTY = "SProperty";
    public static final String S_CONSTRAINT = "SConstraint";
    public static final String ID = "id";
    public static final String VALUE = "value";
    public static final String PARAMETERS = "parameters";
    public static final String S_CONSTRAINT_PARAM = "SConstraintParam";
    public static final String SCHEMA = "Schema";

    private final DataSchema schema;
    private final DataType sConstraintParam;
    private final DataType sConstraint;
    private final DataType sProperty;
    private final DataType sType;
    private final DataType sExtension;
    private final DataType sSchema;
    private final DataFactory factory;

    public SiDOSchema() {
        // Factory
        factory = DataFactories.typedFactory();
        // Creates the schema
        schema = factory.createSchema(SCHEMA);
        // All types
        // SConstraintParam
        sConstraintParam = schema.createDataTypeBuilder(S_CONSTRAINT_PARAM)
                .addStringProperty(NAME)
                .addStringProperty(VALUE).getType();
        // SConstraint
        sConstraint = schema.createDataTypeBuilder(S_CONSTRAINT)
                .addStringProperty(ID)
                .addStringProperty(VALUE, true)
                .addArrayProperty(PARAMETERS, sConstraintParam)
                .getType();
        // SProperty
        sProperty = schema.createDataTypeBuilder(S_PROPERTY)
                .addBooleanProperty(NULLABLE)
                .addBooleanProperty(REVERSE)
                .addBooleanProperty(ARRAY)
                .addStringProperty(ARRAY_INDEX, true)
                .addBooleanProperty(ANONYMOUS)
                .addStringProperty(TYPEREF, true)
                .addStringProperty(NAME)
                .addArrayProperty(CONSTRAINTS, sConstraint)
                .getType();
        // SType
        sType = schema.createDataTypeBuilder(S_TYPE)
                .addStringProperty(NAME)
                .addBooleanProperty(EXTENSIBLE)
                .addBooleanProperty(ABSTRACT)
                .addStringProperty(PARENT, true)
                .addArrayProperty(PROPERTIES, sProperty)
                .getType();
        // SExtension
        sExtension = schema.createDataTypeBuilder(S_EXTENSION)
                .addStringProperty(TYPEREF)
                .addArrayProperty(PROPERTIES, sProperty)
                .getType();
        // SSchema
        sSchema = schema.createDataTypeBuilder(S_SCHEMA)
                .addStringProperty(NAME)
                .addArrayProperty(TYPES, sType)
                .addArrayProperty(EXTENSIONS, sExtension)
                .getType();
        // OK
        schema.close();
    }

    /**
     * Gets the schema of schemas.
     */
    public DataSchema getSchema() {
        return schema;
    }

    public DataObject createSchema(String name) {
        return create(S_SCHEMA)
                .setString(NAME, name);
    }

    private DataObject create(String type) {
        String fullName = SCHEMA + DataSchema.QUALIFIER_SEPARATOR + type;
        return factory.newInstance(fullName, true);
    }

    public DataObject createType(DataObject schema, String name) {
        DataObject type = create(S_TYPE).setString(NAME, name);
        schema.add(TYPES, type);
        return type;
    }

    public DataObject createExtension(DataObject schema, String forType) {
        DataObject extension = create(S_EXTENSION).setString(TYPEREF, forType);
        schema.add(EXTENSIONS, extension);
        return extension;
    }

    public DataObject createProperty(DataObject type) {
        DataObject property = create(S_PROPERTY);
        type.add(PROPERTIES, property);
        return property;
    }

    public DataObject createConstraint(DataObject property, String id) {
        DataObject constraint = create(S_CONSTRAINT).setString(ID, id);
        property.add(CONSTRAINTS, constraint);
        return constraint;
    }

    public void setConstraintParam(DataObject constraint, String name, String value) {
        DataObject param = create(S_CONSTRAINT_PARAM)
                .setString(NAME, name)
                .setString(VALUE, value);
        constraint.add(PARAMETERS, param);
    }
}
